# Copyright 2012-2014 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Tests for the NodeGroup model."""

from __future__ import (
    absolute_import,
    print_function,
    unicode_literals,
    )

str = None

__metaclass__ = type
__all__ = []


from django.db.models.signals import post_save
import django.dispatch
from maasserver.enum import (
    NODEGROUP_STATUS,
    NODEGROUPINTERFACE_MANAGEMENT,
    )
from maasserver.models import (
    BootSource,
    BootSourceSelection,
    Config,
    NodeGroup,
    nodegroup as nodegroup_module,
    )
from maasserver.testing import reload_object
from maasserver.testing.factory import factory
from maasserver.testing.testcase import MAASServerTestCase
from maasserver.worker_user import get_worker_user
from maastesting.celery import CeleryFixture
from mock import (
    call,
    Mock,
    )
from provisioningserver.omshell import generate_omapi_key
from provisioningserver.utils import map_enum
from testresources import FixtureResource
from testtools.matchers import (
    EndsWith,
    GreaterThan,
    HasLength,
    )


class TestNodeGroupManager(MAASServerTestCase):

    def test_new_creates_nodegroup(self):
        name = factory.make_name('nodegroup')
        uuid = factory.make_name('uuid')
        nodegroup = NodeGroup.objects.new(name, uuid)
        nodegroup = reload_object(nodegroup)
        self.assertEqual(name, nodegroup.name)

    def test_new_assigns_token_and_key_for_worker_user(self):
        nodegroup = NodeGroup.objects.new(
            factory.make_name('nodegroup'), factory.make_name('uuid'),
            factory.getRandomIPAddress())
        self.assertIsNotNone(nodegroup.api_token)
        self.assertIsNotNone(nodegroup.api_key)
        self.assertEqual(get_worker_user(), nodegroup.api_token.user)
        self.assertEqual(nodegroup.api_key, nodegroup.api_token.key)

    def test_new_creates_nodegroup_with_empty_dhcp_key(self):
        nodegroup = NodeGroup.objects.new(
            factory.make_name('nodegroup'), factory.make_name('uuid'),
            factory.getRandomIPAddress())
        self.assertEqual('', nodegroup.dhcp_key)

    def test_new_stores_dhcp_key_on_nodegroup(self):
        key = generate_omapi_key()
        nodegroup = NodeGroup.objects.new(
            factory.make_name('nodegroup'), factory.make_name('uuid'),
            factory.getRandomIPAddress(),
            dhcp_key=key)
        self.assertEqual(key, nodegroup.dhcp_key)

    def test_ensure_master_writes_master_nodegroup_to_database(self):
        master = NodeGroup.objects.ensure_master()
        self.assertEqual(
            master.id, NodeGroup.objects.get(name=master.name).id)

    def test_ensure_master_creates_dhcp_key(self):
        master = NodeGroup.objects.ensure_master()
        self.assertThat(len(master.dhcp_key), GreaterThan(20))

    def test_ensure_master_returns_same_nodegroup_every_time(self):
        self.assertEqual(
            NodeGroup.objects.ensure_master().id,
            NodeGroup.objects.ensure_master().id)

    def test_ensure_master_returns_oldest_nodegroup(self):
        first_nodegroup = NodeGroup.objects.new(
            factory.make_name('nodegroup'), factory.make_name('uuid'),
            factory.getRandomIPAddress())
        NodeGroup.objects.new(
            factory.make_name('nodegroup'), factory.make_name('uuid'),
            factory.getRandomIPAddress())
        self.assertEqual(first_nodegroup, NodeGroup.objects.ensure_master())

    def test_ensure_master_preserves_existing_attributes(self):
        master = NodeGroup.objects.ensure_master()
        key = factory.make_string()
        master.dhcp_key = key
        master.save()
        self.assertEqual(key, NodeGroup.objects.ensure_master().dhcp_key)

    def test_ensure_master_creates_accepted_nodegroup(self):
        master = NodeGroup.objects.ensure_master()
        self.assertEqual(NODEGROUP_STATUS.ACCEPTED, master.status)

    def test_get_by_natural_key_looks_up_by_uuid(self):
        nodegroup = factory.make_node_group()
        self.assertEqual(
            nodegroup,
            NodeGroup.objects.get_by_natural_key(nodegroup.uuid))

    def test_get_by_natural_key_will_not_return_other_nodegroup(self):
        factory.make_node_group()
        self.assertRaises(
            NodeGroup.DoesNotExist,
            NodeGroup.objects.get_by_natural_key,
            factory.make_name("nonexistent-nodegroup"))

    def test__mass_change_status_changes_statuses(self):
        old_status = factory.pick_enum(NODEGROUP_STATUS)
        nodegroup1 = factory.make_node_group(status=old_status)
        nodegroup2 = factory.make_node_group(status=old_status)
        new_status = factory.pick_enum(NODEGROUP_STATUS, but_not=[old_status])
        changed = NodeGroup.objects._mass_change_status(old_status, new_status)
        self.assertEqual(
            (
                reload_object(nodegroup1).status,
                reload_object(nodegroup2).status,
                2,
            ),
            (
                new_status,
                new_status,
                changed,
            ))

    def test__mass_change_status_calls_post_save_signal(self):
        old_status = factory.pick_enum(NODEGROUP_STATUS)
        nodegroup = factory.make_node_group(status=old_status)
        recorder = Mock()

        def post_save_NodeGroup(sender, instance, created, **kwargs):
            recorder(instance)

        django.dispatch.Signal.connect(
            post_save, post_save_NodeGroup, sender=NodeGroup)
        self.addCleanup(
            django.dispatch.Signal.disconnect, post_save,
            receiver=post_save_NodeGroup, sender=NodeGroup)
        NodeGroup.objects._mass_change_status(
            old_status, factory.pick_enum(NODEGROUP_STATUS))
        self.assertEqual(
            [call(nodegroup)], recorder.call_args_list)

    def test_reject_all_pending_rejects_nodegroups(self):
        nodegroup = factory.make_node_group(status=NODEGROUP_STATUS.PENDING)
        changed = NodeGroup.objects.reject_all_pending()
        self.assertEqual(
            (NODEGROUP_STATUS.REJECTED, 1),
            (reload_object(nodegroup).status, changed))

    def test_reject_all_pending_does_not_change_others(self):
        unaffected_status = factory.pick_enum(
            NODEGROUP_STATUS, but_not=[NODEGROUP_STATUS.PENDING])
        nodegroup = factory.make_node_group(status=unaffected_status)
        changed_count = NodeGroup.objects.reject_all_pending()
        self.assertEqual(
            (unaffected_status, 0),
            (reload_object(nodegroup).status, changed_count))

    def test_accept_all_pending_accepts_nodegroups(self):
        nodegroup = factory.make_node_group(status=NODEGROUP_STATUS.PENDING)
        changed = NodeGroup.objects.accept_all_pending()
        self.assertEqual(
            (NODEGROUP_STATUS.ACCEPTED, 1),
            (reload_object(nodegroup).status, changed))

    def test_accept_all_pending_does_not_change_others(self):
        unaffected_status = factory.pick_enum(
            NODEGROUP_STATUS, but_not=[NODEGROUP_STATUS.PENDING])
        nodegroup = factory.make_node_group(status=unaffected_status)
        changed_count = NodeGroup.objects.accept_all_pending()
        self.assertEqual(
            (unaffected_status, 0),
            (reload_object(nodegroup).status, changed_count))

    def test_import_boot_images_accepted_clusters_calls_tasks(self):
        recorder = self.patch(nodegroup_module, 'import_boot_images')
        proxy = factory.make_name('proxy')
        Config.objects.set_config('http_proxy', proxy)
        accepted_nodegroups = [
            factory.make_node_group(status=NODEGROUP_STATUS.ACCEPTED),
            factory.make_node_group(status=NODEGROUP_STATUS.ACCEPTED),
        ]
        factory.make_node_group(status=NODEGROUP_STATUS.REJECTED)
        factory.make_node_group(status=NODEGROUP_STATUS.PENDING)
        NodeGroup.objects.import_boot_images_accepted_clusters()
        expected_queues = [
            nodegroup.work_queue
            for nodegroup in accepted_nodegroups]
        actual_queues = [
            kwargs['queue']
            for args, kwargs in recorder.apply_async.call_args_list]
        self.assertItemsEqual(expected_queues, actual_queues)

    def test_refresh_workers_refreshes_accepted_cluster_controllers(self):
        self.patch(nodegroup_module, 'refresh_worker')
        nodegroup = factory.make_node_group(status=NODEGROUP_STATUS.ACCEPTED)
        NodeGroup.objects.refresh_workers()
        nodegroup_module.refresh_worker.assert_called_once_with(nodegroup)

    def test_refresh_workers_skips_unaccepted_cluster_controllers(self):
        self.patch(nodegroup_module, 'refresh_worker')
        for status in map_enum(NODEGROUP_STATUS).values():
            if status != NODEGROUP_STATUS.ACCEPTED:
                factory.make_node_group(status=status)
        NodeGroup.objects.refresh_workers()
        self.assertEqual(0, nodegroup_module.refresh_worker.call_count)


class TestNodeGroup(MAASServerTestCase):

    resources = (
        ('celery', FixtureResource(CeleryFixture())),
        )

    def test_delete_cluster_with_nodes(self):
        nodegroup = factory.make_node_group()
        factory.make_node(nodegroup=nodegroup)
        nodegroup.delete()
        self.assertEqual(nodegroup.uuid, nodegroup.work_queue)
        self.assertFalse(NodeGroup.objects.filter(id=nodegroup.id).exists())

    def test_work_queue_returns_uuid(self):
        nodegroup = factory.make_node_group()
        self.assertEqual(nodegroup.uuid, nodegroup.work_queue)

    def test_manages_dns_returns_True_if_managing_DNS(self):
        nodegroup = factory.make_node_group(
            status=NODEGROUP_STATUS.ACCEPTED,
            management=NODEGROUPINTERFACE_MANAGEMENT.DHCP_AND_DNS)
        self.assertTrue(nodegroup.manages_dns())

    def test_manages_dns_returns_False_if_not_accepted(self):
        nodegroups = [
            factory.make_node_group(
                status=status,
                management=NODEGROUPINTERFACE_MANAGEMENT.DHCP_AND_DNS)
            for status in map_enum(NODEGROUP_STATUS).values()
            ]
        self.assertEqual(
            {
                NODEGROUP_STATUS.PENDING: False,
                NODEGROUP_STATUS.ACCEPTED: True,
                NODEGROUP_STATUS.REJECTED: False,
            },
            {
                nodegroup.status: nodegroup.manages_dns()
                for nodegroup in nodegroups
            })

    def test_manages_dns_returns_False_if_no_interface_manages_DNS(self):
        nodegroups = {
            management: factory.make_node_group(
                status=NODEGROUP_STATUS.ACCEPTED, management=management)
            for management in map_enum(NODEGROUPINTERFACE_MANAGEMENT).values()
            }
        self.assertEqual(
            {
                NODEGROUPINTERFACE_MANAGEMENT.UNMANAGED: False,
                NODEGROUPINTERFACE_MANAGEMENT.DHCP: False,
                NODEGROUPINTERFACE_MANAGEMENT.DHCP_AND_DNS: True,
            },
            {
                management: nodegroup.manages_dns()
                for management, nodegroup in nodegroups.items()
            })

    def test_manages_dns_returns_False_if_nodegroup_has_no_interfaces(self):
        nodegroup = factory.make_node_group(status=NODEGROUP_STATUS.ACCEPTED)
        nodegroup.nodegroupinterface_set.all().delete()
        self.assertFalse(nodegroup.manages_dns())

    def test_get_managed_interfaces_returns_list(self):
        nodegroup = factory.make_node_group()
        self.assertIsInstance(nodegroup.get_managed_interfaces(), list)

    def test_get_managed_interfaces_returns_dhcp_managed_interfaces(self):
        nodegroup = factory.make_node_group(
            management=NODEGROUPINTERFACE_MANAGEMENT.DHCP)
        self.assertEqual(
            set(nodegroup.nodegroupinterface_set.all()),
            set(nodegroup.get_managed_interfaces()))

    def test_get_managed_interfaces_returns_dns_managed_interfaces(self):
        nodegroup = factory.make_node_group(
            management=NODEGROUPINTERFACE_MANAGEMENT.DHCP_AND_DNS)
        self.assertEqual(
            set(nodegroup.nodegroupinterface_set.all()),
            set(nodegroup.get_managed_interfaces()))

    def test_get_managed_interfaces_ignores_unmanaged_interfaces(self):
        nodegroup = factory.make_node_group(
            management=NODEGROUPINTERFACE_MANAGEMENT.UNMANAGED)
        self.assertEqual([], nodegroup.get_managed_interfaces())

    def test_get_managed_interfaces_returns_empty_list_if_none_managed(self):
        nodegroup = factory.make_node_group(
            management=NODEGROUPINTERFACE_MANAGEMENT.UNMANAGED)
        managed_interfaces = nodegroup.get_managed_interfaces()
        self.assertIsInstance(managed_interfaces, list)
        self.assertEqual([], managed_interfaces)

    def test_get_managed_interface_returns_empty_list_if_no_interface(self):
        nodegroup = factory.make_node_group(
            management=NODEGROUPINTERFACE_MANAGEMENT.UNMANAGED)
        nodegroup.nodegroupinterface_set.all().delete()
        self.assertEqual([], nodegroup.get_managed_interfaces())

    def test_accept_node_changes_status(self):
        nodegroup = factory.make_node_group(
            status=factory.pick_enum(NODEGROUP_STATUS))
        nodegroup.accept()
        self.assertEqual(nodegroup.status, NODEGROUP_STATUS.ACCEPTED)

    def test_reject_node_changes_status(self):
        nodegroup = factory.make_node_group(
            status=factory.pick_enum(NODEGROUP_STATUS))
        nodegroup.reject()
        self.assertEqual(nodegroup.status, NODEGROUP_STATUS.REJECTED)

    def test_ensure_dhcp_key_creates_key(self):
        nodegroup = factory.make_node_group(dhcp_key='')
        nodegroup.ensure_dhcp_key()
        # Check that the dhcp_key is not empty and looks
        # valid.
        self.assertThat(nodegroup.dhcp_key, EndsWith("=="))
        # The key is persisted.
        self.assertThat(
            reload_object(nodegroup).dhcp_key, EndsWith("=="))

    def test_ensure_dhcp_key_preserves_existing_key(self):
        key = factory.make_name('dhcp-key')
        nodegroup = factory.make_node_group(dhcp_key=key)
        nodegroup.ensure_dhcp_key()
        self.assertEqual(key, nodegroup.dhcp_key)

    def test_ensure_dhcp_key_creates_different_keys(self):
        nodegroup1 = factory.make_node_group(dhcp_key='')
        nodegroup2 = factory.make_node_group(dhcp_key='')
        nodegroup1.ensure_dhcp_key()
        nodegroup2.ensure_dhcp_key()
        self.assertNotEqual(nodegroup1.dhcp_key, nodegroup2.dhcp_key)

    def test_import_boot_images_calls_import_code_with_proxy(self):
        recorder = self.patch(nodegroup_module, 'import_boot_images')
        self.patch(nodegroup_module, 'report_boot_images', Mock())
        proxy = factory.make_name('proxy')
        Config.objects.set_config('http_proxy', proxy)
        nodegroup = factory.make_node_group()
        nodegroup.import_boot_images()
        [call] = recorder.mock_calls
        _, _, kwargs = call
        self.assertEqual(proxy, kwargs['kwargs'].get('http_proxy'))

    def test_import_boot_images_sent_to_nodegroup_queue(self):
        recorder = self.patch(nodegroup_module, 'import_boot_images')
        nodegroup = factory.make_node_group()
        proxy = factory.make_name('proxy')
        Config.objects.set_config('http_proxy', proxy)
        nodegroup.import_boot_images()
        args, kwargs = recorder.apply_async.call_args
        self.assertEqual(nodegroup.uuid, kwargs['queue'])

    def test_import_boot_images_reports_boot_images(self):
        recorder = self.patch(nodegroup_module, 'import_boot_images')
        report_recorder = self.patch(
            nodegroup_module, 'report_boot_images')
        nodegroup = factory.make_node_group()
        nodegroup.import_boot_images()

        # The 'import_boot_images' subtask is called with queue=nodegroup's
        # UUID.  This means the task will be executed by the cluster's celery.
        self.assertEqual(
            dict(options={'queue': nodegroup.uuid}),
            report_recorder.subtask.call_args[1],
        )
        self.assertEqual(
            report_recorder.subtask(),
            recorder.apply_async.call_args[1]['kwargs']['callback'],
        )

    def test_import_boot_images_passes_sources_list(self):
        recorder = self.patch(nodegroup_module, 'import_boot_images')
        self.patch(nodegroup_module, 'report_boot_images', Mock())
        nodegroup = factory.make_node_group()
        nodegroup.ensure_boot_source_definition()
        sources = [
            source.to_dict() for source in nodegroup.bootsource_set.all()]
        nodegroup.import_boot_images()
        [call] = recorder.mock_calls
        _, _, kwargs = call
        self.assertEqual(sources, kwargs['kwargs'].get('sources'))

    def test_ensure_boot_source_definition_creates_default_source(self):
        cluster = factory.make_node_group()
        cluster.ensure_boot_source_definition()
        sources = BootSource.objects.filter(cluster=cluster)
        self.assertThat(sources, HasLength(1))
        [source] = sources
        self.assertAttributes(
            source,
            {
                'url': 'http://maas.ubuntu.com/images/ephemeral-v2/releases/',
                'keyring_filename': (
                    '/usr/share/keyrings/ubuntu-cloudimage-keyring.gpg'),
            })
        selections = BootSourceSelection.objects.filter(boot_source=source)
        by_release = {
            selection.release: selection
            for selection in selections
            }
        self.assertItemsEqual(['precise', 'trusty'], by_release.keys())
        self.assertAttributes(
            by_release['precise'],
            {
                'release': 'precise',
                'arches': ['*'],
                'subarches': ['*'],
                'labels': ['release'],
            })
        self.assertAttributes(
            by_release['trusty'],
            {
                'release': 'trusty',
                'arches': ['*'],
                'subarches': ['*'],
                'labels': ['release'],
            })

    def test_ensure_boot_source_definition_skips_if_already_present(self):
        cluster = factory.make_node_group()
        sources = [
            factory.make_boot_source(cluster=cluster)
            for _ in range(3)
            ]
        cluster.ensure_boot_source_definition()
        self.assertItemsEqual(sources, cluster.bootsource_set.all())
